import asyncio
import logging
from itertools import (takewhile,
                       islice,
                       tee)
from typing import List, Iterator, Tuple, Dict, Generator, Iterable, Union

from arps.core.clock import TimeEvent
from arps.core.simulator.event_queue_loader import EventQueueLoader
from arps.core.simulator.sim_event import SimEvent
from arps.core.simulator.sim_event_scheduler import SimEventScheduler


class Simulator:
    '''
    Discrete event simulator
    '''

    def __init__(self, event_queue_loader: Union[Iterable, EventQueueLoader],
                 sim_event_scheduler: SimEventScheduler) -> None:
        '''
        Args:
        - event_queue_loader : event queue, not format defined yet
        - sim_event_scheduler: event scheduler
        '''
        self._has_events = True
        self._event_queue_loader = event_queue_loader
        self._event_gen = iter(self._event_queue_loader)
        self._current_events: List[SimEvent] = list()
        self._event_scheduler = sim_event_scheduler
        self.logger = logging.getLogger(self.__class__.__name__)
        self._current_coro_events: Dict[int, Generator] = dict()
        self._finished_events: 'asyncio.Queue[SimEvent]' = asyncio.Queue()

    @property
    def finished(self):
        return not self._has_events

    def step(self, time_event: TimeEvent) -> None:
        '''
        Process events that are in the current epoch. Only process events if the epoch
        changed

        Return if there are more events in the event queue
        '''
        current_time = time_event.value
        self.logger.debug('running step at epoch time %s', current_time)
        events, self._event_gen, events_available = self.next_events(self._event_gen, current_time)
        self._current_events.extend(events)

        events_scheduled = self._event_scheduler.schedule(self._current_events, current_time)

        for event in events_scheduled:
            self.logger.debug('processing events at epoch time %s', current_time)
            if event.identifier not in self._current_coro_events:
                event_process = event.process()
                event_process.send(None)
                self._current_coro_events[event.identifier] = event_process

        for event_task_id, event_coro in self._current_coro_events.items():
            try:
                event_coro.send(current_time)
            except StopIteration:
                pass

        for event in self._current_events:
            if event.finished_time is not None:
                self._finished_events.put_nowait(event)
                del self._current_coro_events[event.identifier]

        self._current_events = [event for event in
                                self._current_events if event.finished_time is None]

        self._has_events = events_available or len(self._current_events) > 0

    def reset(self):
        '''
        Reset simulator attributes needed to run simulator from scratch

        '''
        self._event_gen = iter(self._event_queue_loader)
        self._has_events = True

    def next_events(self, events_gen, current_time: int) -> Tuple[List[SimEvent], Iterator[SimEvent], bool]:
        '''
        Returns events that arrived at the current epoch,
        if there are more events, and updated generator of events

        Args:
        - events_gen: current generator of events
        - current_time: current epoch
        '''
        current_events_it, next_event_it, has_next_it = tee(events_gen, 3)

        def current_event_predicate(event):
            return event.arrival_time == current_time

        current_events = list(takewhile(current_event_predicate, current_events_it))
        event_gen = islice(next_event_it, len(current_events), None)
        has_more = True
        try:
            next(has_next_it)
        except StopIteration:
            has_more = False

        return current_events, event_gen, has_more
